from autoarray.plot import plotters
from autoarray.structures import arrays
from autogalaxy.plot import lensing_plotters, light_profile_plots, mass_profile_plots


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def image(galaxy, grid, positions=None, include=None, plotter=None):
    """Plot the image (e.g. the datas) of a galaxy, on a grid of (y,x) coordinates.

    Set *autogalaxy.datas.arrays.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.ag.Galaxy
        The galaxy whose image are plotted.
    grid : grid_like or datas.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    plotter.plot_array(
        array=galaxy.image_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=galaxy),
        light_profile_centres=include.light_profile_centres_from_obj(obj=galaxy),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=galaxy),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def convergence(galaxy, grid, positions=None, include=None, plotter=None):
    """Plot the convergence of a galaxy, on a grid of (y,x) coordinates.

    Set *autogalaxy.datas.arrays.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.ag.Galaxy
        The galaxy whose convergence is plotted.
    grid : grid_like or datas.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    plotter.plot_array(
        array=galaxy.convergence_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=galaxy),
        light_profile_centres=include.light_profile_centres_from_obj(obj=galaxy),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=galaxy),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def potential(galaxy, grid, positions=None, include=None, plotter=None):
    """Plot the potential of a galaxy, on a grid of (y,x) coordinates.

     Set *autogalaxy.datas.arrays.plotters.plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.ag.Galaxy
         The galaxy whose potential is plotted.
    grid : grid_like or datas.arrays.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
     """
    plotter.plot_array(
        array=galaxy.potential_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=galaxy),
        light_profile_centres=include.light_profile_centres_from_obj(obj=galaxy),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=galaxy),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def deflections_y(galaxy, grid, positions=None, include=None, plotter=None):
    """Plot the y component of the deflection angles of a galaxy, on a grid of (y,x) coordinates.

    Set *autogalaxy.datas.arrays.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.ag.Galaxy
        The galaxy whose y deflecton angles are plotted.
    grid : grid_like or datas.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    deflections = galaxy.deflections_from_grid(grid=grid)
    deflections_y = arrays.Array.manual_mask(
        array=deflections.in_1d[:, 0], mask=grid.mask
    )

    plotter.plot_array(
        array=deflections_y,
        mask=include.mask_from_grid(grid=grid),
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=galaxy),
        light_profile_centres=include.light_profile_centres_from_obj(obj=galaxy),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=galaxy),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def deflections_x(galaxy, grid, positions=None, include=None, plotter=None):
    """Plot the x component of the deflection angles of a galaxy, on a grid of (y,x) coordinates.

     Set *autogalaxy.datas.arrays.plotters.plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.ag.Galaxy
         The galaxy whose x deflecton angles are plotted.
     grid : grid_like or datas.arrays.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
     """
    deflections = galaxy.deflections_from_grid(grid=grid)
    deflections_x = arrays.Array.manual_mask(
        array=deflections.in_1d[:, 1], mask=grid.mask
    )
    plotter.plot_array(
        array=deflections_x,
        mask=include.mask_from_grid(grid=grid),
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=galaxy),
        light_profile_centres=include.light_profile_centres_from_obj(obj=galaxy),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=galaxy),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def magnification(galaxy, grid, positions=None, include=None, plotter=None):
    """Plot the magnification of a galaxy, on a grid of (y,x) coordinates.

     Set *autogalaxy.datas.arrays.plotters.plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.ag.Galaxy
         The galaxy whose magnification is plotted.
    grid : grid_like or datas.arrays.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
     """

    plotter.plot_array(
        array=galaxy.magnification_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=galaxy),
        light_profile_centres=include.light_profile_centres_from_obj(obj=galaxy),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=galaxy),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_labels
def image_subplot(galaxy, grid, positions=None, include=None, sub_plotter=None):

    number_subplots = len(galaxy.light_profiles)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for i, light_profile in enumerate(galaxy.light_profiles):

        sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=i + 1)

        light_profile_plots.image(
            light_profile=light_profile,
            grid=grid,
            positions=positions,
            include=include,
            plotter=sub_plotter,
        )

    sub_plotter.output.subplot_to_figure()
    sub_plotter.figure.close()


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_labels
def convergence_subplot(galaxy, grid, positions=None, include=None, sub_plotter=None):

    number_subplots = len(galaxy.mass_profiles)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=i + 1)

        mass_profile_plots.convergence(
            mass_profile=mass_profile,
            grid=grid,
            positions=positions,
            include=include,
            plotter=sub_plotter,
        )

    sub_plotter.output.subplot_to_figure()
    sub_plotter.figure.close()


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_labels
def potential_subplot(galaxy, grid, positions=None, include=None, sub_plotter=None):

    number_subplots = len(galaxy.mass_profiles)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=i + 1)

        mass_profile_plots.potential(
            mass_profile=mass_profile,
            grid=grid,
            positions=positions,
            include=include,
            plotter=sub_plotter,
        )

    sub_plotter.output.subplot_to_figure()
    sub_plotter.figure.close()


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_labels
def deflections_y_subplot(galaxy, grid, positions=None, include=None, sub_plotter=None):

    number_subplots = len(galaxy.mass_profiles)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=i + 1)

        mass_profile_plots.deflections_y(
            mass_profile=mass_profile,
            grid=grid,
            positions=positions,
            include=include,
            plotter=sub_plotter,
        )

    sub_plotter.output.subplot_to_figure()
    sub_plotter.figure.close()


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_labels
def deflections_x_subplot(galaxy, grid, positions=None, include=None, sub_plotter=None):

    number_subplots = len(galaxy.mass_profiles)

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=i + 1)

        mass_profile_plots.deflections_x(
            mass_profile=mass_profile,
            grid=grid,
            positions=positions,
            include=include,
            plotter=sub_plotter,
        )

    sub_plotter.output.subplot_to_figure()
    sub_plotter.figure.close()


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def contribution_map(galaxy, mask=None, positions=None, include=None, plotter=None):
    """Plot the summed contribution maps of a hyper_galaxies-fit.

    Set *autogalaxy.datas.arrays.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractLensHyperFit
        The hyper_galaxies-fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the contribution_maps are plotted.
    """

    plotter.plot_array(
        array=galaxy.contribution_map,
        mask=mask,
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=galaxy),
        light_profile_centres=include.light_profile_centres_from_obj(obj=galaxy),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=galaxy),
        include_origin=include.origin,
    )



from autogalaxy.dataset import interferometer
from autogalaxy.pipeline.phase.dataset import meta_dataset


class MetaInterferometer(meta_dataset.MetaDataset):
    def __init__(self, settings, model, real_space_mask, is_hyper_phase=False):
        super().__init__(settings=settings, model=model, is_hyper_phase=is_hyper_phase)
        self.real_space_mask = real_space_mask

    def masked_dataset_from(self, dataset, mask, results, modified_visibilities):

        real_space_mask = self.mask_with_phase_sub_size_from_mask(
            mask=self.real_space_mask
        )

        masked_interferometer = interferometer.MaskedInterferometer(
            interferometer=dataset.modified_visibilities_from_visibilities(
                modified_visibilities
            ),
            visibilities_mask=mask,
            real_space_mask=real_space_mask,
            grid_class=self.settings.grid_class,
            grid_inversion_class=self.settings.grid_inversion_class,
            fractional_accuracy=self.settings.fractional_accuracy,
            sub_steps=self.settings.sub_steps,
            transformer_class=self.settings.transformer_class,
            primary_beam_shape_2d=self.settings.primary_beam_shape_2d,
            inversion_pixel_limit=self.settings.inversion_pixel_limit,
        )

        return masked_interferometer



from __future__ import division, print_function

import os

from autoconf import conf
import autogalaxy as ag

import numpy as np
import pytest

directory = os.path.dirname(os.path.realpath(__file__))

grid_10 = ag.Grid.manual_2d(grid=np.array([[[1.0, 0.0]]]), pixel_scales=1.0, sub_size=1)
grid_zero = ag.Grid.manual_2d(
    grid=np.array([[[0.0000000001, 0.0]]]), pixel_scales=1.0, sub_size=1
)


class TestGaussian:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        gaussian = ag.lp.EllipticalGaussian(centre=(0.0, 0.0))

        image_1 = gaussian.image_from_grid(grid=grid_10)
        image_0 = gaussian.image_from_grid(grid=grid_zero)

        assert image_0 == pytest.approx(image_1, 1.0e-4)

        gaussian = ag.lp.SphericalGaussian(centre=(0.0, 0.0))

        image_1 = gaussian.image_from_grid(grid=grid_10)
        image_0 = gaussian.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestSersic:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        sersic = ag.lp.EllipticalSersic(centre=(0.0, 0.0))

        image_1 = sersic.image_from_grid(grid=grid_10)
        image_0 = sersic.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        sersic = ag.lp.SphericalSersic(centre=(0.0, 0.0))

        image_1 = sersic.image_from_grid(grid=grid_10)
        image_0 = sersic.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestExponential:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        exponential = ag.lp.EllipticalExponential(centre=(0.0, 0.0))

        image_1 = exponential.image_from_grid(grid=grid_10)
        image_0 = exponential.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        exponential = ag.lp.SphericalExponential(centre=(0.0, 0.0))

        image_1 = exponential.image_from_grid(grid=grid_10)
        image_0 = exponential.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestDevVaucouleurs:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        dev_vaucouleurs = ag.lp.EllipticalDevVaucouleurs(centre=(0.0, 0.0))

        image_1 = dev_vaucouleurs.image_from_grid(grid=grid_10)
        image_0 = dev_vaucouleurs.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        dev_vaucouleurs = ag.lp.SphericalDevVaucouleurs(centre=(0.0, 0.0))

        image_1 = dev_vaucouleurs.image_from_grid(grid=grid_10)
        image_0 = dev_vaucouleurs.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestCoreSersic:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        core_sersic = ag.lp.EllipticalCoreSersic(centre=(0.0, 0.0))

        image_1 = core_sersic.image_from_grid(grid=grid_10)
        image_0 = core_sersic.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        core_sersic = ag.lp.SphericalCoreSersic(centre=(0.0, 0.0))

        image_1 = core_sersic.image_from_grid(grid=grid_10)
        image_0 = core_sersic.image_from_grid(grid=grid_zero)
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestPointMass:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        point_mass = ag.mp.PointMass(centre=(0.0, 0.0), einstein_radius=1.0)

        deflections_1 = point_mass.deflections_from_grid(grid=grid_10)
        deflections_0 = point_mass.deflections_from_grid(
            grid=np.array([[0.00000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestCoredPowerLaw:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        cored_power_law = ag.mp.EllipticalCoredPowerLaw(
            centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0
        )

        convergence_1 = cored_power_law.convergence_from_grid(grid=grid_10)
        convergence_0 = cored_power_law.convergence_from_grid(
            grid=np.array([[1e-8, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)


class TestPowerLaw:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        power_law = ag.mp.EllipticalPowerLaw(
            centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0
        )

        convergence_1 = power_law.convergence_from_grid(grid=grid_10)
        convergence_0 = power_law.convergence_from_grid(grid=np.array([[1.0e-9, 0.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)


class TestCoredIsothermal:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        cored_isothermal = ag.mp.EllipticalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        convergence_1 = cored_isothermal.convergence_from_grid(grid=grid_10)
        convergence_0 = cored_isothermal.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        cored_isothermal = ag.mp.SphericalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        convergence_1 = cored_isothermal.convergence_from_grid(grid=grid_10)
        convergence_0 = cored_isothermal.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        cored_isothermal = ag.mp.EllipticalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        potential_1 = cored_isothermal.potential_from_grid(grid=grid_10)
        potential_0 = cored_isothermal.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        cored_isothermal = ag.mp.SphericalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        potential_1 = cored_isothermal.potential_from_grid(grid=grid_10)
        potential_0 = cored_isothermal.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        cored_isothermal = ag.mp.EllipticalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        deflections_1 = cored_isothermal.deflections_from_grid(grid=grid_10)
        deflections_0 = cored_isothermal.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        cored_isothermal = ag.mp.SphericalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        deflections_1 = cored_isothermal.deflections_from_grid(grid=grid_10)
        deflections_0 = cored_isothermal.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestIsothermal:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        isothermal = ag.mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

        convergence_1 = isothermal.convergence_from_grid(grid=grid_10)
        convergence_0 = isothermal.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        isothermal = ag.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

        convergence_1 = isothermal.convergence_from_grid(grid=grid_10)
        convergence_0 = isothermal.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        isothermal = ag.mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

        potential_1 = isothermal.potential_from_grid(grid=grid_10)
        potential_0 = isothermal.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        isothermal = ag.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

        potential_1 = isothermal.potential_from_grid(grid=grid_10)
        potential_0 = isothermal.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        isothermal = ag.mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

        deflections_1 = isothermal.deflections_from_grid(grid=grid_10)
        deflections_0 = isothermal.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        isothermal = ag.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

        deflections_1 = isothermal.deflections_from_grid(grid=grid_10)
        deflections_0 = isothermal.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestGeneralizedNFW:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        gnfw = ag.mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))

        convergence_1 = gnfw.convergence_from_grid(grid=grid_10)
        convergence_0 = gnfw.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        gnfw = ag.mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))

        potential_1 = gnfw.potential_from_grid(grid=grid_10)
        potential_0 = gnfw.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        gnfw = ag.mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))

        deflections_1 = gnfw.deflections_from_grid(grid=grid_10)
        deflections_0 = gnfw.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestTruncatedNFW:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        truncated_nfw = ag.mp.SphericalTruncatedNFW(centre=(0.0, 0.0))

        convergence_1 = truncated_nfw.convergence_from_grid(grid=grid_10)
        convergence_0 = truncated_nfw.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        truncated_nfw = ag.mp.SphericalTruncatedNFW(centre=(0.0, 0.0))

        potential_1 = truncated_nfw.potential_from_grid(grid=grid_10)
        potential_0 = truncated_nfw.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        truncated_nfw = ag.mp.SphericalTruncatedNFW(centre=(0.0, 0.0))

        deflections_1 = truncated_nfw.deflections_from_grid(grid=grid_10)
        deflections_0 = truncated_nfw.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestNFW:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        nfw = ag.mp.EllipticalNFW(centre=(0.0, 0.0))

        convergence_1 = nfw.convergence_from_grid(grid=grid_10)
        convergence_0 = nfw.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        nfw = ag.mp.SphericalNFW(centre=(0.0, 0.0))

        convergence_1 = nfw.convergence_from_grid(grid=grid_10)
        convergence_0 = nfw.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        nfw = ag.mp.EllipticalNFW(centre=(0.0, 0.0))

        potential_1 = nfw.potential_from_grid(grid=grid_10)
        potential_0 = nfw.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        nfw = ag.mp.SphericalNFW(centre=(0.0, 0.0))

        potential_1 = nfw.potential_from_grid(grid=grid_10)
        potential_0 = nfw.potential_from_grid(grid=grid_zero)
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        nfw = ag.mp.EllipticalNFW(centre=(0.0, 0.0))

        deflections_1 = nfw.deflections_from_grid(grid=grid_10)
        deflections_0 = nfw.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        nfw = ag.mp.SphericalNFW(centre=(0.0, 0.0))

        deflections_1 = nfw.deflections_from_grid(grid=grid_10)
        deflections_0 = nfw.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestSersicMass:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        sersic = ag.mp.EllipticalSersic(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_10)
        convergence_0 = sersic.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = ag.mp.SphericalSersic(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_10)
        convergence_0 = sersic.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = ag.mp.EllipticalSersic(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_10)
        deflections_0 = sersic.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = ag.mp.SphericalSersic(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_10)
        deflections_0 = sersic.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestExponentialMass:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        exponential = ag.mp.EllipticalExponential(centre=(0.0, 0.0))

        convergence_1 = exponential.convergence_from_grid(grid=grid_10)
        convergence_0 = exponential.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        exponential = ag.mp.SphericalExponential(centre=(0.0, 0.0))

        convergence_1 = exponential.convergence_from_grid(grid=grid_10)
        convergence_0 = exponential.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        exponential = ag.mp.EllipticalExponential(centre=(0.0, 0.0))

        deflections_1 = exponential.deflections_from_grid(grid=grid_10)
        deflections_0 = exponential.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        exponential = ag.mp.SphericalExponential(centre=(0.0, 0.0))

        deflections_1 = exponential.deflections_from_grid(grid=grid_10)
        deflections_0 = exponential.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestDevVaucouleursMass:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        dev_vaucouleurs = ag.mp.EllipticalDevVaucouleurs(centre=(0.0, 0.0))

        convergence_1 = dev_vaucouleurs.convergence_from_grid(grid=grid_10)
        convergence_0 = dev_vaucouleurs.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        dev_vaucouleurs = ag.mp.EllipticalDevVaucouleurs(centre=(0.0, 0.0))

        deflections_1 = dev_vaucouleurs.deflections_from_grid(grid=grid_10)
        deflections_0 = dev_vaucouleurs.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        dev_vaucouleurs = ag.mp.SphericalDevVaucouleurs(centre=(0.0, 0.0))

        deflections_1 = dev_vaucouleurs.deflections_from_grid(grid=grid_10)
        deflections_0 = dev_vaucouleurs.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestSersicMassRadialGradient:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        sersic = ag.mp.EllipticalSersicRadialGradient(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_10)
        convergence_0 = sersic.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = ag.mp.SphericalSersicRadialGradient(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_10)
        convergence_0 = sersic.convergence_from_grid(grid=grid_zero)
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = ag.mp.EllipticalSersicRadialGradient(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_10)
        deflections_0 = sersic.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = ag.mp.SphericalSersicRadialGradient(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_10)
        deflections_0 = sersic.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestMassSheet:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0))

        deflections_1 = mass_sheet.deflections_from_grid(grid=grid_10)
        deflections_0 = mass_sheet.deflections_from_grid(grid=grid_zero)
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestExternalShear:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        conf.instance = conf.Config(config_path="{}/files/config/".format(directory))

        shear = ag.mp.ExternalShear(elliptical_comps=(0.1, 0.1))

        deflections_1 = shear.deflections_from_grid(grid=np.array([[1e-8, 0.0]]))
        deflections_0 = shear.deflections_from_grid(grid=np.array([[1e-9, 0.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)



import autofit as af
import autogalaxy as ag
from autogalaxy import exc

import pytest


def test__pixelization__model_depends_on_inversion_pixels_fixed():

    setup = ag.PipelineSetup()

    assert setup.pixelization is None

    setup = ag.PipelineSetup(pixelization=ag.pix.Rectangular)

    assert setup.pixelization is ag.pix.Rectangular

    setup = ag.PipelineSetup(pixelization=ag.pix.VoronoiBrightnessImage)

    assert setup.pixelization is ag.pix.VoronoiBrightnessImage

    setup = ag.PipelineSetup(
        pixelization=ag.pix.VoronoiBrightnessImage, inversion_pixels_fixed=100
    )

    assert isinstance(setup.pixelization, af.PriorModel)
    assert setup.pixelization.pixels == 100


def test__hyper_searches():

    setup = ag.PipelineSetup(hyper_galaxies=False)
    assert setup.hyper_galaxies_search == None

    setup = ag.PipelineSetup(hyper_galaxies=True)
    assert setup.hyper_galaxies_search.n_live_points == 75
    assert setup.hyper_galaxies_search.evidence_tolerance == pytest.approx(
        0.084, 1.0e-4
    )

    setup = ag.PipelineSetup(
        hyper_galaxies=True, hyper_galaxies_search=af.DynestyStatic(n_live_points=51)
    )
    assert setup.hyper_galaxies_search.n_live_points == 51
    assert setup.hyper_galaxies_search.evidence_tolerance == pytest.approx(0.06, 1.0e-4)

    setup = ag.PipelineSetup(inversion_search=None)
    assert setup.inversion_search.n_live_points == 50
    assert setup.inversion_search.evidence_tolerance == pytest.approx(0.059, 1.0e-4)

    setup = ag.PipelineSetup(inversion_search=af.DynestyStatic(n_live_points=51))
    assert setup.inversion_search.n_live_points == 51

    setup = ag.PipelineSetup(hyper_combined_search=af.DynestyStatic(n_live_points=51))
    assert setup.hyper_combined_search.n_live_points == 51

    setup = ag.PipelineSetup(hyper_galaxies=True, evidence_tolerance=0.5)
    assert setup.hyper_galaxies_search.evidence_tolerance == pytest.approx(
        0.084, 1.0e-4
    )
    assert setup.inversion_search.evidence_tolerance == 0.5
    assert setup.hyper_combined_search.evidence_tolerance == 0.5

    with pytest.raises(exc.PipelineException):

        ag.PipelineSetup(
            hyper_galaxies=True,
            hyper_galaxies_search=af.DynestyStatic(n_live_points=51),
            evidence_tolerance=0.5,
        )

    with pytest.raises(exc.PipelineException):
        ag.PipelineSetup(
            inversion_search=af.DynestyStatic(n_live_points=51), evidence_tolerance=3.0
        )

    with pytest.raises(exc.PipelineException):
        ag.PipelineSetup(
            hyper_combined_search=af.DynestyStatic(n_live_points=51),
            evidence_tolerance=3.0,
        )


def test__hyper_galaxies_tag():

    setup = ag.PipelineSetup(hyper_galaxies=False)
    assert setup.hyper_galaxies_tag == ""

    setup = ag.PipelineSetup(hyper_galaxies=True)
    assert setup.hyper_galaxies_tag == "_galaxies"


def test__hyper_image_sky_tag():
    setup = ag.PipelineSetup(hyper_image_sky=False)
    assert setup.hyper_galaxies_tag == ""

    setup = ag.PipelineSetup(hyper_image_sky=True)
    assert setup.hyper_image_sky_tag == "_bg_sky"


def test__hyper_background_noise_tag():
    setup = ag.PipelineSetup(hyper_background_noise=False)
    assert setup.hyper_galaxies_tag == ""

    setup = ag.PipelineSetup(hyper_background_noise=True)
    assert setup.hyper_background_noise_tag == "_bg_noise"


def test__hyper_tag():

    setup = ag.PipelineSetup(
        hyper_galaxies=True, hyper_image_sky=True, hyper_background_noise=True
    )

    assert setup.hyper_tag == "__hyper_galaxies_bg_sky_bg_noise"

    setup = ag.PipelineSetup(hyper_galaxies=True, hyper_background_noise=True)

    assert setup.hyper_tag == "__hyper_galaxies_bg_noise"


def test__pixelization_tag():
    setup = ag.PipelineSetup(pixelization=None)
    assert setup.pixelization_tag == ""
    setup = ag.PipelineSetup(pixelization=ag.pix.Rectangular)
    assert setup.pixelization_tag == "pix_rect"
    setup = ag.PipelineSetup(pixelization=ag.pix.VoronoiBrightnessImage)
    assert setup.pixelization_tag == "pix_voro_image"


def test__regularization_tag():
    setup = ag.PipelineSetup(regularization=None)
    assert setup.regularization_tag == ""
    setup = ag.PipelineSetup(regularization=ag.reg.Constant)
    assert setup.regularization_tag == "__reg_const"
    setup = ag.PipelineSetup(regularization=ag.reg.AdaptiveBrightness)
    assert setup.regularization_tag == "__reg_adapt_bright"


def test__inversion_pixels_fixed_tag():

    setup = ag.PipelineSetup(inversion_pixels_fixed=None)
    assert setup.inversion_pixels_fixed_tag == ""

    setup = ag.PipelineSetup(inversion_pixels_fixed=100)
    assert setup.inversion_pixels_fixed_tag == ""

    setup = ag.PipelineSetup(
        inversion_pixels_fixed=100, pixelization=ag.pix.VoronoiBrightnessImage
    )
    assert setup.inversion_pixels_fixed_tag == "_100"


def test__inversion_tag():

    setup = ag.PipelineSetup(pixelization=None, inversion_pixels_fixed=100)
    assert setup.inversion_tag == ""
    setup = ag.PipelineSetup(regularization=None, inversion_pixels_fixed=100)
    assert setup.inversion_tag == ""
    setup = ag.PipelineSetup(
        pixelization=ag.pix.Rectangular,
        regularization=ag.reg.Constant,
        inversion_pixels_fixed=100,
    )
    assert setup.inversion_tag == "__pix_rect__reg_const"
    setup = ag.PipelineSetup(
        pixelization=ag.pix.VoronoiBrightnessImage,
        regularization=ag.reg.AdaptiveBrightness,
        inversion_pixels_fixed=None,
    )
    assert setup.inversion_tag == "__pix_voro_image__reg_adapt_bright"

    setup = ag.PipelineSetup(
        pixelization=ag.pix.VoronoiBrightnessImage,
        regularization=ag.reg.AdaptiveBrightness,
        inversion_pixels_fixed=100,
    )
    assert setup.inversion_tag == "__pix_voro_image_100__reg_adapt_bright"


def test__light_centre_tag():

    setup = ag.PipelineSetup(light_centre=None)
    assert setup.light_centre_tag == ""
    setup = ag.PipelineSetup(light_centre=(2.0, 2.0))
    assert setup.light_centre_tag == "__light_centre_(2.00,2.00)"
    setup = ag.PipelineSetup(light_centre=(3.0, 4.0))
    assert setup.light_centre_tag == "__light_centre_(3.00,4.00)"
    setup = ag.PipelineSetup(light_centre=(3.027, 4.033))
    assert setup.light_centre_tag == "__light_centre_(3.03,4.03)"


def test__align_bulge_disk_tags():

    light = ag.PipelineSetup(align_bulge_disk_centre=False)
    assert light.align_bulge_disk_centre_tag == ""
    light = ag.PipelineSetup(align_bulge_disk_centre=True)
    assert light.align_bulge_disk_centre_tag == "_centre"

    light = ag.PipelineSetup(align_bulge_disk_elliptical_comps=False)
    assert light.align_bulge_disk_elliptical_comps_tag == ""
    light = ag.PipelineSetup(align_bulge_disk_elliptical_comps=True)
    assert light.align_bulge_disk_elliptical_comps_tag == "_ell"


def test__bulge_disk_tag():
    light = ag.PipelineSetup(
        align_bulge_disk_centre=False, align_bulge_disk_elliptical_comps=False
    )
    assert light.align_bulge_disk_tag == ""

    light = ag.PipelineSetup(
        align_bulge_disk_centre=True, align_bulge_disk_elliptical_comps=False
    )
    print(light.align_bulge_disk_tag)
    assert light.align_bulge_disk_tag == "__align_bulge_disk_centre"

    light = ag.PipelineSetup(
        align_bulge_disk_centre=True, align_bulge_disk_elliptical_comps=True
    )
    assert light.align_bulge_disk_tag == "__align_bulge_disk_centre_ell"


def test__disk_as_sersic_tag():
    light = ag.PipelineSetup(disk_as_sersic=False)
    assert light.disk_as_sersic_tag == ""
    light = ag.PipelineSetup(disk_as_sersic=True)
    assert light.disk_as_sersic_tag == "__disk_sersic"


def test__number_of_gaussians_tag():
    setup = ag.PipelineSetup()
    assert setup.number_of_gaussians_tag == ""
    setup = ag.PipelineSetup(number_of_gaussians=1)
    assert setup.number_of_gaussians_tag == "__gaussians_x1"
    setup = ag.PipelineSetup(number_of_gaussians=2)
    assert setup.number_of_gaussians_tag == "__gaussians_x2"


def test__tag():

    setup = ag.PipelineSetup(
        pixelization=ag.pix.Rectangular,
        regularization=ag.reg.Constant,
        light_centre=(1.0, 2.0),
    )

    setup.type_tag = setup.inversion_tag

    assert setup.tag == "setup__pix_rect__reg_const__light_centre_(1.00,2.00)"

    setup = ag.PipelineSetup(number_of_gaussians=1)

    assert setup.tag == "setup__gaussians_x1"



import autofit as af
import numpy as np
from astropy import cosmology as cosmo
from autoarray.operators.inversion import inversions as inv
from autoarray.structures import arrays, grids, visibilities as vis
from autogalaxy import dimensions as dim
from autogalaxy import exc
from autogalaxy import lensing
from autogalaxy.galaxy import galaxy as g
from autogalaxy.util import cosmology_util
from autogalaxy.util import plane_util


class AbstractPlane(lensing.LensingObject):
    def __init__(self, redshift, galaxies, cosmology):
        """A plane of galaxies where all galaxies are at the same redshift.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of galaxies in this plane.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        if redshift is None:

            if not galaxies:
                raise exc.PlaneException(
                    "No redshift and no galaxies were input to a Plane. A redshift for the Plane therefore cannot be"
                    "determined"
                )
            elif not all(
                [galaxies[0].redshift == galaxy.redshift for galaxy in galaxies]
            ):
                redshift = np.mean([galaxy.redshift for galaxy in galaxies])
            else:
                redshift = galaxies[0].redshift

        self.redshift = redshift
        self.galaxies = galaxies
        self.cosmology = cosmology

    @property
    def galaxy_redshifts(self):
        return [galaxy.redshift for galaxy in self.galaxies]

    @property
    def has_light_profile(self):
        if self.galaxies is not None:
            return any(
                list(map(lambda galaxy: galaxy.has_light_profile, self.galaxies))
            )

    @property
    def has_mass_profile(self):
        if self.galaxies is not None:
            return any(list(map(lambda galaxy: galaxy.has_mass_profile, self.galaxies)))

    @property
    def has_pixelization(self):
        return any([galaxy.pixelization for galaxy in self.galaxies])

    @property
    def has_regularization(self):
        return any([galaxy.regularization for galaxy in self.galaxies])

    @property
    def galaxies_with_light_profile(self):
        return list(filter(lambda galaxy: galaxy.has_light_profile, self.galaxies))

    @property
    def galaxies_with_mass_profile(self):
        return list(filter(lambda galaxy: galaxy.has_mass_profile, self.galaxies))

    @property
    def galaxies_with_pixelization(self):
        return list(filter(lambda galaxy: galaxy.has_pixelization, self.galaxies))

    @property
    def galaxies_with_regularization(self):
        return list(filter(lambda galaxy: galaxy.has_regularization, self.galaxies))

    @property
    def pixelization(self):

        if len(self.galaxies_with_pixelization) == 0:
            return None
        if len(self.galaxies_with_pixelization) == 1:
            return self.galaxies_with_pixelization[0].pixelization
        elif len(self.galaxies_with_pixelization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with pixelizations in one plane is above 1"
            )

    @property
    def regularization(self):

        if len(self.galaxies_with_regularization) == 0:
            return None
        if len(self.galaxies_with_regularization) == 1:
            return self.galaxies_with_regularization[0].regularization
        elif len(self.galaxies_with_regularization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with regularizations in one plane is above 1"
            )

    @property
    def hyper_galaxy_image_of_galaxy_with_pixelization(self):
        galaxies_with_pixelization = self.galaxies_with_pixelization
        if galaxies_with_pixelization:
            return galaxies_with_pixelization[0].hyper_galaxy_image

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda galaxy: galaxy.has_hyper_galaxy, self.galaxies)))

    @property
    def light_profile_centres(self):
        """Returns the light profile centres of the plane as a *GridCoordinates* object, which structures the centres
        in lists according to which galaxy they come from.

        Fo example, if a plane has two galaxies, the first with one light profile and second with two light profiles
        this  returns:

        [[(y0, x0)], [(y0, x0), (y1, x1)]]
        
        This is used for visualization, for example plotting the centres of all light profiles colored by their galaxy.
        """
        return grids.GridCoordinates(
            [
                list(galaxy.light_profile_centres)
                for galaxy in self.galaxies
                if galaxy.has_light_profile
            ]
        )

    @property
    def mass_profiles(self):
        return [
            item
            for mass_profile in self.mass_profiles_of_galaxies
            for item in mass_profile
        ]

    @property
    def mass_profiles_of_galaxies(self):
        return [
            galaxy.mass_profiles for galaxy in self.galaxies if galaxy.has_mass_profile
        ]

    @property
    def mass_profile_centres(self):
        """Returns the mass profile centres of the plane as a *GridCoordinates* object, which structures the centres
        in lists according to which galaxy they come from.

        Fo example, if a plane has two galaxies, the first with one mass profile and second with two mass profiles
        this  returns:

        [[(y0, x0)], [(y0, x0), (y1, x1)]]

        This is used for visualization, for example plotting the centres of all mass profiles colored by their galaxy.

        The centres of mass-sheets are filtered out, as their centres are not relevant to lensing calculations.
        """
        return grids.GridCoordinates(
            [
                list(galaxy.mass_profile_centres)
                for galaxy in self.galaxies
                if galaxy.has_mass_profile and not galaxy.has_only_mass_sheets
            ]
        )

    @property
    def mass_profile_axis_ratios(self):
        """Returns the mass profile axis-ratios of the plane as a *GridCoordinates* object, which structures the axis-ratios
        in lists according to which galaxy they come from.

        Fo example, if a plane has two galaxies, the first with one mass profile and second with two mass profiles
        this  returns:

        [[axis_ratio_0], [axis_ratio_0, axis_ratio_1]]

        This is used for visualization, for example plotting ellipses of axis-ratios of all mass profiles colored by 
        their galaxy.
        """
        return arrays.Values(
            [
                list(galaxy.mass_profile_axis_ratios)
                for galaxy in self.galaxies
                if galaxy.has_mass_profile
            ]
        )

    @property
    def mass_profile_phis(self):
        """Returns the mass profile phis of the plane as a *GridCoordinates* object, which structures the phis
        in lists according to which galaxy they come from.

        Fo example, if a plane has two galaxies, the first with one mass profile and second with two mass profiles
        this  returns:

        [[phi_0], [phi_0, phi_1]]

        This is used for visualization, for example plotting ellipses of phis of all mass profiles colored by 
        their galaxy.
        """
        return arrays.Values(
            [
                list(galaxy.mass_profile_phis)
                for galaxy in self.galaxies
                if galaxy.has_mass_profile
            ]
        )

    def new_object_with_units_converted(
        self,
        unit_length=None,
        unit_luminosity=None,
        unit_mass=None,
        kpc_per_arcsec=None,
        exposure_time=None,
        critical_surface_density=None,
    ):

        new_galaxies = list(
            map(
                lambda galaxy: galaxy.new_object_with_units_converted(
                    unit_length=unit_length,
                    unit_luminosity=unit_luminosity,
                    unit_mass=unit_mass,
                    kpc_per_arcsec=kpc_per_arcsec,
                    exposure_time=exposure_time,
                    critical_surface_density=critical_surface_density,
                ),
                self.galaxies,
            )
        )

        return self.__class__(
            galaxies=new_galaxies, redshift=self.redshift, cosmology=self.cosmology
        )

    @property
    def unit_length(self):
        if self.has_light_profile:
            return self.galaxies_with_light_profile[0].unit_length
        elif self.has_mass_profile:
            return self.galaxies_with_mass_profile[0].unit_length
        else:
            return None

    @property
    def unit_luminosity(self):
        if self.has_light_profile:
            return self.galaxies_with_light_profile[0].unit_luminosity
        elif self.has_mass_profile:
            return self.galaxies_with_mass_profile[0].unit_luminosity
        else:
            return None

    @property
    def unit_mass(self):
        if self.has_mass_profile:
            return self.galaxies_with_mass_profile[0].unit_mass
        else:
            return None


class AbstractPlaneCosmology(AbstractPlane):
    def __init__(self, redshift, galaxies, cosmology):

        super(AbstractPlaneCosmology, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    @property
    def arcsec_per_kpc(self):
        return cosmology_util.arcsec_per_kpc_from(
            redshift=self.redshift, cosmology=self.cosmology
        )

    @property
    def kpc_per_arcsec(self):
        return 1.0 / self.arcsec_per_kpc

    def angular_diameter_distance_to_earth_in_units(self, unit_length="arcsec"):
        return cosmology_util.angular_diameter_distance_to_earth_from(
            redshift=self.redshift, cosmology=self.cosmology, unit_length=unit_length
        )

    def cosmic_average_density_in_units(
        self, unit_length="arcsec", unit_mass="angular"
    ):
        return cosmology_util.cosmic_average_density_from(
            redshift=self.redshift,
            cosmology=self.cosmology,
            unit_length=unit_length,
            unit_mass=unit_mass,
        )


class AbstractPlaneLensing(AbstractPlaneCosmology):
    def __init__(self, redshift, galaxies, cosmology):
        super(AbstractPlaneCosmology, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    @grids.grid_like_to_structure
    def image_from_grid(self, grid):
        """Compute the profile-image plane image of the list of galaxies of the plane's sub-grid, by summing the
        individual images of each galaxy's light profile.

        The image is calculated on the sub-grid and binned-up to the original grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an arrays of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------

        """
        if self.galaxies:
            return sum(
                map(lambda galaxy: galaxy.image_from_grid(grid=grid), self.galaxies)
            )
        return np.zeros((grid.shape[0],))

    def images_of_galaxies_from_grid(self, grid):
        return list(
            map(lambda galaxy: galaxy.image_from_grid(grid=grid), self.galaxies)
        )

    def padded_image_from_grid_and_psf_shape(self, grid, psf_shape_2d):

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_2d=psf_shape_2d)

        return self.image_from_grid(grid=padded_grid)

    @grids.grid_like_to_structure
    def convergence_from_grid(self, grid):
        """Compute the convergence of the list of galaxies of the plane's sub-grid, by summing the individual convergences \
        of each galaxy's mass profile.

        The convergence is calculated on the sub-grid and binned-up to the original grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an arrays of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        grid : Grid
            The grid (or sub) of (y,x) arc-second coordinates at the centre of every unmasked pixel which the \
            potential is calculated on.
        galaxies : [g.Galaxy]
            The galaxies whose mass profiles are used to compute the surface densities.
        """
        if self.galaxies:
            return sum(map(lambda g: g.convergence_from_grid(grid=grid), self.galaxies))
        else:
            return np.zeros(shape=(grid.shape[0],))

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        """Compute the potential of the list of galaxies of the plane's sub-grid, by summing the individual potentials \
        of each galaxy's mass profile.

        The potential is calculated on the sub-grid and binned-up to the original grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an arrays of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        grid : Grid
            The grid (or sub) of (y,x) arc-second coordinates at the centre of every unmasked pixel which the \
            potential is calculated on.
        galaxies : [g.Galaxy]
            The galaxies whose mass profiles are used to compute the surface densities.
        """
        if self.galaxies:
            return sum(map(lambda g: g.potential_from_grid(grid=grid), self.galaxies))
        return np.zeros((grid.shape[0]))

    @grids.grid_like_to_structure
    def deflections_from_grid(self, grid):
        if self.galaxies:
            return sum(map(lambda g: g.deflections_from_grid(grid=grid), self.galaxies))
        return np.zeros(shape=(grid.shape[0], 2))

    @grids.grid_like_to_structure
    def traced_grid_from_grid(self, grid):
        """Trace this plane's grid_stacks to the next plane, using its deflection angles."""

        return grid - self.deflections_from_grid(grid=grid)

    def luminosities_of_galaxies_within_circles_in_units(
        self, radius: dim.Length, unit_luminosity="eps", exposure_time=None
    ):
        """Compute the total luminosity of all galaxies in this plane within a circle of specified radius.

        See *galaxy.light_within_circle* and *light_profiles.light_within_circle* for details \
        of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_luminosity : str
            The unit_label the luminosity is returned in {esp, counts}.
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second unit_label to counts.
        """
        return list(
            map(
                lambda galaxy: galaxy.luminosity_within_circle_in_units(
                    radius=radius,
                    unit_luminosity=unit_luminosity,
                    exposure_time=exposure_time,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def masses_of_galaxies_within_circles_in_units(
        self, radius: dim.Length, unit_mass="angular", redshift_source=None
    ):
        """Compute the total mass of all galaxies in this plane within a circle of specified radius.

        See *galaxy.angular_mass_within_circle* and *mass_profiles.angular_mass_within_circle* for details
        of how this is performed.

        Parameters
        ----------
        redshift_source
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_mass : str
            The unit_label the mass is returned in {angular, angular}.

        """
        return list(
            map(
                lambda galaxy: galaxy.mass_within_circle_in_units(
                    radius=radius,
                    unit_mass=unit_mass,
                    redshift_source=redshift_source,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )


class AbstractPlaneData(AbstractPlaneLensing):
    def __init__(self, redshift, galaxies, cosmology):

        super(AbstractPlaneData, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    def blurred_image_from_grid_and_psf(self, grid, psf, blurring_grid):

        image = self.image_from_grid(grid=grid)

        blurring_image = self.image_from_grid(grid=blurring_grid)

        return psf.convolved_array_from_array_2d_and_mask(
            array_2d=image.in_2d_binned + blurring_image.in_2d_binned, mask=grid.mask
        )

    def blurred_images_of_galaxies_from_grid_and_psf(self, grid, psf, blurring_grid):
        return [
            galaxy.blurred_image_from_grid_and_psf(
                grid=grid, psf=psf, blurring_grid=blurring_grid
            )
            for galaxy in self.galaxies
        ]

    def blurred_image_from_grid_and_convolver(self, grid, convolver, blurring_grid):

        image = self.image_from_grid(grid=grid)

        blurring_image = self.image_from_grid(grid=blurring_grid)

        return convolver.convolved_image_from_image_and_blurring_image(
            image=image, blurring_image=blurring_image
        )

    def blurred_images_of_galaxies_from_grid_and_convolver(
        self, grid, convolver, blurring_grid
    ):
        return [
            galaxy.blurred_image_from_grid_and_convolver(
                grid=grid, convolver=convolver, blurring_grid=blurring_grid
            )
            for galaxy in self.galaxies
        ]

    def unmasked_blurred_image_from_grid_and_psf(self, grid, psf):

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_2d=psf.shape_2d)

        padded_image = self.image_from_grid(grid=padded_grid)

        return padded_grid.mask.unmasked_blurred_array_from_padded_array_psf_and_image_shape(
            padded_array=padded_image, psf=psf, image_shape=grid.mask.shape
        )

    def unmasked_blurred_image_of_galaxies_from_grid_and_psf(self, grid, psf):

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_2d=psf.shape_2d)

        unmasked_blurred_images_of_galaxies = []

        for galaxy in self.galaxies:

            padded_image_1d = galaxy.image_from_grid(grid=padded_grid)

            unmasked_blurred_array_2d = padded_grid.mask.unmasked_blurred_array_from_padded_array_psf_and_image_shape(
                padded_array=padded_image_1d, psf=psf, image_shape=grid.mask.shape
            )

            unmasked_blurred_images_of_galaxies.append(unmasked_blurred_array_2d)

        return unmasked_blurred_images_of_galaxies

    def profile_visibilities_from_grid_and_transformer(self, grid, transformer):

        if self.galaxies:
            image = self.image_from_grid(grid=grid)
            return transformer.visibilities_from_image(image=image)
        else:
            return vis.Visibilities.zeros(
                shape_1d=(transformer.uv_wavelengths.shape[0],)
            )

    def profile_visibilities_of_galaxies_from_grid_and_transformer(
        self, grid, transformer
    ):
        return [
            galaxy.profile_visibilities_from_grid_and_transformer(
                grid=grid, transformer=transformer
            )
            for galaxy in self.galaxies
        ]

    def sparse_image_plane_grid_from_grid(self, grid, inversion_stochastic=False):

        if not self.has_pixelization:
            return None

        hyper_galaxy_image = self.hyper_galaxy_image_of_galaxy_with_pixelization

        return self.pixelization.sparse_grid_from_grid(
            grid=grid,
            hyper_image=hyper_galaxy_image,
            inversion_stochastic=inversion_stochastic,
        )

    def mapper_from_grid_and_sparse_grid(
        self, grid, sparse_grid, inversion_uses_border=False
    ):

        galaxies_with_pixelization = list(
            filter(lambda galaxy: galaxy.pixelization is not None, self.galaxies)
        )

        if len(galaxies_with_pixelization) == 0:
            return None
        if len(galaxies_with_pixelization) == 1:

            pixelization = galaxies_with_pixelization[0].pixelization

            return pixelization.mapper_from_grid_and_sparse_grid(
                grid=grid,
                sparse_grid=sparse_grid,
                inversion_uses_border=inversion_uses_border,
                hyper_image=galaxies_with_pixelization[0].hyper_galaxy_image,
            )

        elif len(galaxies_with_pixelization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with pixelizations in one plane is above 1"
            )

    def inversion_imaging_from_grid_and_data(
        self, grid, image, noise_map, convolver, inversion_uses_border=False
    ):

        sparse_grid = self.sparse_image_plane_grid_from_grid(grid=grid)

        mapper = self.mapper_from_grid_and_sparse_grid(
            grid=grid,
            sparse_grid=sparse_grid,
            inversion_uses_border=inversion_uses_border,
        )

        return inv.InversionImaging.from_data_mapper_and_regularization(
            image=image,
            noise_map=noise_map,
            convolver=convolver,
            mapper=mapper,
            regularization=self.regularization,
        )

    def inversion_interferometer_from_grid_and_data(
        self,
        grid,
        visibilities,
        noise_map,
        transformer,
        visibilities_complex=None,
        inversion_uses_border=False,
    ):

        sparse_grid = self.sparse_image_plane_grid_from_grid(grid=grid)

        mapper = self.mapper_from_grid_and_sparse_grid(
            grid=grid,
            sparse_grid=sparse_grid,
            inversion_uses_border=inversion_uses_border,
        )

        return inv.InversionInterferometer.from_data_mapper_and_regularization(
            visibilities=visibilities,
            noise_map=noise_map,
            transformer=transformer,
            visibilities_complex=visibilities_complex,
            mapper=mapper,
            regularization=self.regularization,
        )

    def plane_image_from_grid(self, grid):
        return plane_util.plane_image_of_galaxies_from(
            shape=grid.mask.shape,
            grid=grid.geometry.unmasked_grid,
            galaxies=self.galaxies,
        )

    def hyper_noise_map_from_noise_map(self, noise_map):
        hyper_noise_maps = self.hyper_noise_maps_of_galaxies_from_noise_map(
            noise_map=noise_map
        )
        return sum(hyper_noise_maps)

    def hyper_noise_maps_of_galaxies_from_noise_map(self, noise_map):
        """For a contribution map and noise-map, use the model hyper_galaxy galaxies to compute a hyper noise-map.

        Parameters
        -----------
        noise_map : imaging.NoiseMap or ndarray
            An arrays describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        """
        hyper_noise_maps = []

        for galaxy in self.galaxies:
            if galaxy.has_hyper_galaxy:

                hyper_noise_map_1d = galaxy.hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                    noise_map=noise_map,
                    hyper_model_image=galaxy.hyper_model_image,
                    hyper_galaxy_image=galaxy.hyper_galaxy_image,
                )

                hyper_noise_maps.append(hyper_noise_map_1d)

            else:

                hyper_noise_map = arrays.Array.manual_mask(
                    array=np.zeros(noise_map.mask.mask_sub_1.pixels_in_mask),
                    mask=noise_map.mask.mask_sub_1,
                )

                hyper_noise_maps.append(hyper_noise_map)

        return hyper_noise_maps

    @property
    def contribution_map(self):

        contribution_maps = self.contribution_maps_of_galaxies
        if None in contribution_maps:
            contribution_maps = [i for i in contribution_maps if i is not None]

        if contribution_maps:
            return sum(contribution_maps)
        else:
            return None

    @property
    def contribution_maps_of_galaxies(self):

        contribution_maps = []

        for galaxy in self.galaxies:

            if galaxy.hyper_galaxy is not None:

                contribution_maps.append(galaxy.contribution_map)

            else:

                contribution_maps.append(None)

        return contribution_maps

    def galaxy_image_dict_from_grid(self, grid) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_image_dict = dict()

        images_of_galaxies = self.images_of_galaxies_from_grid(grid=grid)
        for (galaxy_index, galaxy) in enumerate(self.galaxies):
            galaxy_image_dict[galaxy] = images_of_galaxies[galaxy_index]

        return galaxy_image_dict

    def galaxy_blurred_image_dict_from_grid_and_convolver(
        self, grid, convolver, blurring_grid
    ) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_blurred_image_dict = dict()

        blurred_images_of_galaxies = self.blurred_images_of_galaxies_from_grid_and_convolver(
            grid=grid, convolver=convolver, blurring_grid=blurring_grid
        )
        for (galaxy_index, galaxy) in enumerate(self.galaxies):
            galaxy_blurred_image_dict[galaxy] = blurred_images_of_galaxies[galaxy_index]

        return galaxy_blurred_image_dict

    def galaxy_profile_visibilities_dict_from_grid_and_transformer(
        self, grid, transformer
    ) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_profile_visibilities_image_dict = dict()

        profile_visibilities_of_galaxies = self.profile_visibilities_of_galaxies_from_grid_and_transformer(
            grid=grid, transformer=transformer
        )
        for (galaxy_index, galaxy) in enumerate(self.galaxies):
            galaxy_profile_visibilities_image_dict[
                galaxy
            ] = profile_visibilities_of_galaxies[galaxy_index]

        return galaxy_profile_visibilities_image_dict

    @property
    def yticks(self):
        """Compute the yticks labels of this grid, used for plotting the y-axis ticks when visualizing an image \
        """
        return np.linspace(np.amin(self.grid[:, 0]), np.amax(self.grid[:, 0]), 4)

    @property
    def xticks(self):
        """Compute the xticks labels of this grid, used for plotting the x-axis ticks when visualizing an \
        image"""
        return np.linspace(np.amin(self.grid[:, 1]), np.amax(self.grid[:, 1]), 4)


class Plane(AbstractPlaneData):
    def __init__(self, redshift=None, galaxies=None, cosmology=cosmo.Planck15):

        super(Plane, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    # noinspection PyUnusedLocal
    def summarize_in_units(
        self,
        radii,
        whitespace=80,
        unit_length="arcsec",
        unit_luminosity="eps",
        unit_mass="angular",
        redshift_source=None,
        **kwargs,
    ):

        summary = ["Plane\n"]
        prefix_plane = ""

        summary += [
            af.formatter.label_and_value_string(
                label=prefix_plane + "redshift",
                value=self.redshift,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        summary += [
            af.formatter.label_and_value_string(
                label=prefix_plane + "kpc_per_arcsec",
                value=self.kpc_per_arcsec,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        angular_diameter_distance_to_earth = self.angular_diameter_distance_to_earth_in_units(
            unit_length=unit_length
        )

        summary += [
            af.formatter.label_and_value_string(
                label=prefix_plane + "angular_diameter_distance_to_earth",
                value=angular_diameter_distance_to_earth,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        for galaxy in self.galaxies:
            summary += ["\n"]
            summary += galaxy.summarize_in_units(
                radii=radii,
                whitespace=whitespace,
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                unit_mass=unit_mass,
                redshift_source=redshift_source,
                cosmology=self.cosmology,
            )

        return summary


class PlaneImage:
    def __init__(self, array, grid):
        self.array = array
        self.grid = grid

    @property
    def xticks(self):
        return self.array.mask.geometry.xticks

    @property
    def yticks(self):
        return self.array.mask.geometry.yticks
